import pytest
from ms_ovba_compression.ms_ovba import MsOvba


ceil_log2_data = [
    (1, 4),
    (2, 4),
    (3, 4),
    (4, 4),
    (9, 4),
    (17, 5),
    (50, 6),
]


@pytest.mark.parametrize("input, expected", ceil_log2_data)
def test_ceil_log2(input: int, expected: int) -> None:
    assert MsOvba.ceil_log2(input) == expected


pack_copytoken_data = [
    ([3, 8, 9], 0x7000),
    ([4, 8, 17], 0x3801),
    ([3, 7, 24], 0x3000),
    ([5, 15, 32], 0x7002),
    ([7, 5, 38], 0x1004),
    ([3, 16, 54], 0x3C00),
]


@pytest.mark.parametrize("inputs, expected", pack_copytoken_data)
def test_pack_copytoken(inputs: list, expected: int) -> None:
    help = MsOvba.copytoken_help(inputs[2])
    assert MsOvba.pack_copytoken(inputs[0], inputs[1], help) == expected


copytoken_help_data = [
    (9, [4, 0x0FFF, 0xF000]),
    (17, [5, 0x07FF, 0xF800]),
]


@pytest.mark.parametrize("input, expected", copytoken_help_data)
def test_copytoken_help(input: int, expected: list) -> None:
    result = MsOvba.copytoken_help(input)
    assert result["bitCount"] == expected[0]
    assert result["lengthMask"] == expected[1]
    assert result["offsetMask"] == expected[2]


unpack_copytoken_data = [
    ([9, 0x7000], [3, 8]),
    ([32, 0x7002], [5, 15]),
    ([54, 0x3C00], [3, 16]),
]


@pytest.mark.parametrize("inputs, expected", unpack_copytoken_data)
def test_unpack_copytoken(inputs: list, expected: list) -> None:
    help = MsOvba.copytoken_help(inputs[0])
    result = MsOvba.unpack_copytoken(inputs[1], help)
    assert result["length"] == expected[0]
    assert result["offset"] == expected[1]



import pytest
from ms_ovba_compression.ms_ovba import MsOvba


def test_normal_compression() -> None:
    ms_ovba = MsOvba()
    expected = b"#aaabcdefaaaaghijaaaaaklaaamnopqaaaaaaaaaaaarstuvwxyzaaa"
    compressed = (b'\x01\x2F\xB0\x00\x23\x61\x61\x61\x62\x63\x64\x65\x82\x66'
                  + b'\x00\x70\x61\x67\x68\x69\x6A\x01\x38\x08\x61\x6B\x6C\x00'
                  + b'\x30\x6D\x6E\x6F\x70\x06\x71\x02\x70\x04\x10\x72\x73\x74'
                  + b'\x75\x76\x10\x77\x78\x79\x7A\x00\x3C')
    assert ms_ovba.decompress(compressed) == expected


def test_bad_signature_byte() -> None:
    """
    The container must have a signature byte
    """
    ms_ovba = MsOvba()
    compressed = (b'\x2F\xB0\x00\x23\x61\x61\x61\x62\x63\x64\x65\x82\x66\x00'
                  + b'\x70\x61\x67\x68\x69\x6A\x01\x38\x08\x61\x6B\x6C\x00\x30'
                  + b'\x6D\x6E\x6F\x70\x06\x71\x02\x70\x04\x10\x72\x73\x74\x75'
                  + b'\x76\x10\x77\x78\x79\x7A\x00\x3C')
    with pytest.raises(Exception):
        ms_ovba.decompress(compressed)


def test_container_too_short() -> None:
    """
    The container cannot be shorter then specified by the header
    """
    ms_ovba = MsOvba()
    compressed = (b'\x01\x2F\xB0\x00\x23\x61\x61\x61\x62\x63\x64\x65\x82\x66'
                  + b'\x00\x70\x61\x67\x68\x69\x6A\x01\x38\x08\x61\x6B\x6C\x00'
                  + b'\x30\x6D\x6E\x6F\x70\x06\x71\x02\x70\x04\x10\x72\x73\x74'
                  + b'\x75\x76\x10\x77\x78\x79\x7A\x00')
    with pytest.raises(Exception):
        ms_ovba.decompress(compressed)


def test_container_too_long() -> None:
    """
    The container must not be longer then specified in the header
    """
    ms_ovba = MsOvba()
    compressed = (b'\x01\x2F\xB0\x00\x23\x61\x61\x61\x62\x63\x64\x65\x82\x66'
                  + b'\x00\x70\x61\x67\x68\x69\x6A\x01\x38\x08\x61\x6B\x6C\x00'
                  + b'\x30\x6D\x6E\x6F\x70\x06\x71\x02\x70\x04\x10\x72\x73\x74'
                  + b'\x75\x76\x10\x77\x78\x79\x7A\x00\3C\x3F\xFF')
    with pytest.raises(Exception):
        ms_ovba.decompress(compressed)


def test_bad_raw_chunk_header() -> None:
    """
    Raw chunks must have 4096 bytes of data. Header = 0x3FFF
    """
    ms_ovba = MsOvba()
    compressed = (b'\x01\xFE\x3F\x00\x23\x61\x61\x61\x62\x63\x64\x65\x82\x66'
                  + b'\x00\x70\x61\x67\x68\x69\x6A\x01\x38\x08\x61\x6B\x6C\x00'
                  + b'\x30\x6D\x6E\x6F\x70\x06\x71\x02\x70\x04\x10\x72\x73\x74'
                  + b'\x75\x76\x10\x77\x78\x79\x7A\x00')
    with pytest.raises(Exception):
        ms_ovba.decompress(compressed)


def test_flag_with_no_data() -> None:
    """
    If a flag byte is present, there must be data after it.
    """
    ms_ovba = MsOvba()
    compressed = (b'\x01\x29\xB0\x00\x23\x61\x61\x61\x62\x63\x64\x65\x82\x66'
                  + b'\x00\x70\x61\x67\x68\x69\x6A\x01\x38\x08\x61\x6B\x6C\x00'
                  + b'\x30\x6D\x6E\x6F\x70\x06\x71\x02\x70\x04\x10\x72\x73\x74'
                  + b'\x75\x76\x10')
    with pytest.raises(Exception):
        ms_ovba.decompress(compressed)


def test_bad_header_signature() -> None:
    """
    The header signature must be 0b011
    """
    ms_ovba = MsOvba()
    compressed = (b'\x01\x2F\xA0\x00\x23\x61\x61\x61\x62\x63\x64\x65\x82\x66'
                  + b'\x00\x70\x61\x67\x68\x69\x6A\x01\x38\x08\x61\x6B\x6C\x00'
                  + b'\x30\x6D\x6E\x6F\x70\x06\x71\x02\x70\x04\x10\x72\x73\x74'
                  + b'\x75\x76\x10\x77\x78\x79\x7A\x00\x3C')
    with pytest.raises(Exception):
        ms_ovba.decompress(compressed)


def test_incomplete_copytoken() -> None:
    """
    A copy token must be two bytes. test for early termination
    """
    ms_ovba = MsOvba()
    compressed = (b'\x01\x2A\xB0\x00\x23\x61\x61\x61\x62\x63\x64\x65\x82\x66'
                  + b'\x00\x70\x61\x67\x68\x69\x6A\x01\x38\x08\x61\x6B\x6C\x00'
                  + b'\x30\x6D\x6E\x6F\x70\x06\x71\x02\x70\x04\x10\x72\x73\x74'
                  + b'\x75\x76\x01\x77')
    with pytest.raises(Exception):
        ms_ovba.decompress(compressed)


def test_decompress_unable_to_compress_big() -> None:
    expected = b'abcdefghijklmnopqrstuv.'
    ms_ovba = MsOvba("big")
    input = (b'\x01\xB0\x19\x00\x61\x62\x63\x64\x65\x66\x67\x68\x00\x69\x6A'
             + b'\x6B\x6C\x6D\x6E\x6F\x70\x00\x71\x72\x73\x74\x75\x76\x2E')
    assert ms_ovba.decompress(input) == expected



from ms_ovba_compression.ms_ovba import MsOvba


def test_unable_to_compress() -> None:
    input = b'abcdefghijklmnopqrstuv.'
    ms_ovba = MsOvba()
    expected = (b'\x01\x19\xB0\x00\x61\x62\x63\x64\x65\x66\x67\x68\x00\x69\x6A'
                + b'\x6B\x6C\x6D\x6E\x6F\x70\x00\x71\x72\x73\x74\x75\x76\x2E')
    assert ms_ovba.compress(input) == expected


def test_max_compression() -> None:
    input = (b'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
             + b'aaaaaaaaaa')
    ms_ovba = MsOvba()
    expected = b'\x01\x03\xB0\x02\x61\x45\x00'
    assert ms_ovba.compress(input) == expected


def test_long_poor_compression() -> None:
    """
    Every sequence of 8 bytes has a flag byte prepended to the compressed token
    sequence. In theory a 3640 byte sequence could "compress" to be larger than
    4096 bytes.
    """
    data = b''
    for i in range(7):
        for j in range(256):
            data += i.to_bytes(1, "little") + j.to_bytes(1, "little")
    i = 8
    for j in range(72):
        data += i.to_bytes(1, "little") + j.to_bytes(1, "little")
    assert len(data) < 4096
    ms_ovba = MsOvba()
    result = ms_ovba.compress(data)
    # the length is 1 container signature byte, 2 header bytes, and the
    # original data padded to 4096 bytes
    assert len(result) == 4099
    # The resulting chunk header will be 0x3FFF
    assert result[2] & 0xF0 == 0x30


def test_unable_to_compress_big() -> None:
    input = b'abcdefghijklmnopqrstuv.'
    ms_ovba = MsOvba("big")
    expected = (b'\x01\xB0\x19\x00\x61\x62\x63\x64\x65\x66\x67\x68\x00\x69\x6A'
                + b'\x6B\x6C\x6D\x6E\x6F\x70\x00\x71\x72\x73\x74\x75\x76\x2E')
    assert ms_ovba.compress(input) == expected



from typing import Literal, TypeVar


T = TypeVar('T', bound='MsOvba')


class MsOvba:

    def __init__(self: T, endian: Literal['little', 'big'] = 'little') -> None:
        self._endian = endian

    def decompress(self: T, compressed_container: bytes) -> bytes:
        """
        Decompress a compressed container usng the MS-OVBA Compression
        Algorithm

        :param compressedContainer bytes: bytes of compressed data
        :return: Uncompressed data
        :rtype: bytes
        """
        uncompressed_data = b''
        # The compressed container must begin with the byte \x01
        if compressed_container[0] != 0x01:
            message = ("The container signature byte must be \\x01, not "
                       + str(compressed_container[0]) + ".")
            raise Exception(message)
        # Pop off the signature byte. Everything else is compressed chunks
        chunks = compressed_container[1:]
        while len(chunks) > 0:
            # The first two bytes of each chunk is the header. It will tell us
            # how long the compressed data is in this chunk. All chunks must be
            # 4096 bytes uncompressed except for the last chunk.
            header = chunks[0:2]
            compressed, length = self._unpack_header(header)

            # The unpackHeader method gives us the chunk length. The data
            # portion is two less than that.
            compressed_data_length = length - 2

            # If we have less data then we are supposed to, we have a problem.
            if len(chunks) < length:
                message = ("Expecting " + str(length - 2)
                           + " data bytes, but given " + str(len(chunks) - 2)
                           + ".")
                raise Exception(message)

            # Split out the compresseddata from the chunk buffer.
            compressed_chunk = chunks[2:compressed_data_length + 2]

            # Pop off the data we are working on from the buffer
            chunks = chunks[length + 2:]
            decompressed_chunk = self._decompress_chunk(compressed_chunk)
            uncompressed_data += decompressed_chunk

            # If the last chunk is less than 4096 bytes, there better not be
            # anything left in the buffer. Should this raise a warning instead?
            if len(decompressed_chunk) < 4096 and len(chunks) > 0:
                message = "The provided compressed container is too long."
                raise Exception(message)
        return uncompressed_data

    def _unpack_header(self: T, compressed_header: bytes) -> tuple:
        # Need to find out if this bit order is endian dependent. It seems the
        # real world data had the bits packed little endian and then the
        # resulting two bytes packed little endian into the binary file.
        int_header = int.from_bytes(compressed_header, self._endian)

        # Data is compressed if the least significat bit is 0b1
        compressed = (int_header & 0x8000) >> 15

        # the 12 most significant bits is three less than the chunk size
        length = (int_header & 0x0FFF) + 3
        if compressed == 0 and length != 4098:
            raise Exception("If uncompressed, chunk must be 4096 bytes.")
        signature = (int_header & 0x7000) >> 12
        if signature != 3:
            message = ("Chunk signature must be three. Value is "
                       + str(signature) + ".")
            raise Exception(message)
        return compressed, length

    def _decompress_chunk(self: T, compressed_chunk: bytes) -> bytes:
        """
        Decompress bytes object

        :param data bytes: bytes of compressed data
        :return: bytes
        :rtype: bytes
        """
        uncompressed_chunk = b''
        while len(compressed_chunk) > 0:
            # The Flag Byte is one byte. pop it off
            flag_byte = compressed_chunk[0]
            compressed_chunk = compressed_chunk[1:]

            # If we have a flag byte, we better have data to go with it.
            if len(compressed_chunk) == 0:
                message = ("There must be at least one token "
                           "in each TokenSequence.")
                raise Exception(message)
            flag_mask = 1
            for i in range(8):
                # Extract Flag bit from the token with the mask
                flag_bit = flag_byte & flag_mask

                if flag_bit == 0:
                    # If the flag bit is zero, no compression ocurred, so just
                    # move the byte over.
                    if len(compressed_chunk) > 0:
                        byte = compressed_chunk[0].to_bytes(1, "little")
                        uncompressed_chunk += byte
                        compressed_chunk = compressed_chunk[1:]
                else:
                    # If the flag bit is one, grab the 2 byte copy token and
                    # determine the offset and length of the replacement
                    # string. There better be 2 bytes or we're in trouble.
                    if len(compressed_chunk) < 2:
                        message = "Copy Token does not exist."
                        raise Exception(message)
                    help = MsOvba.copytoken_help(len(uncompressed_chunk))
                    # The copy Token is always packed into the compressed chuck
                    # little endian.
                    copytoken = int.from_bytes(compressed_chunk[:2], "little")
                    copytoken_data = MsOvba.unpack_copytoken(copytoken, help)
                    compressed_chunk = compressed_chunk[2:]
                    offset = copytoken_data["offset"]

                    # Copy data from the uncompressed chunk, {offset} bytes
                    # away, {length} number of times. Note that this can mean
                    # that we could possibly copy new data multiple times, ie.
                    # offset 1 length 7
                    for i in range(copytoken_data["length"]):
                        copy_int = uncompressed_chunk[-1 * offset]
                        copy_byte = copy_int.to_bytes(1, "little")
                        uncompressed_chunk += copy_byte
                # Move the mask for the next round
                flag_mask = flag_mask << 1
        return uncompressed_chunk

    def compress(self: T, data: bytes) -> bytes:
        """
        Compress a bytearray
        :param data bytes: bytes of compressed data
        :return: compressed data
        :rtype: bytes
        """
        # The compressed container begins with a sgnature byte.
        compressed_container = b'\x01'

        # Each chunk must hold 4096 bytes of data except the final chunk.
        number_of_chunks = (len(data) - 1) // 4096 + 1

        # Grab 4096 bytes, compress the chunk, and add it to the container
        for i in range(number_of_chunks):
            start = i * 4096
            end = (i + 1) * 4096
            compressed_chunk = self._compress_chunk(data[start: end])
            compressed_container += compressed_chunk

        return compressed_container

    def _compress_chunk(self: T, data: bytes) -> bytes:
        """
        A chunk of data is 4096 bytes or less. This will return a stream of max
        length 4098, a 2 byte header and up to 4096 bytes of data.
        """

        # A record of the uncompressed data is needed after compression for the
        # matching algorithm.
        self._activeChunk = data

        # Initialize with an empty header
        # Is endian-ness supposed to affect the header organization?
        # The docs state 12 length bits + 0b011 + compression-bit but real
        # world little endian file has compression-bit + 0b011 + 12 length bits
        compress_and_sig = 0xB000

        # Buffer of data to be compressed and moved to compressedChunk
        self._uncompressedData = data

        # Buffer of compressed data.
        compressed_chunk = b''

        # While there is data in the uncompressed buffer, build a token
        # sequence and compress it and add the compressed sequence to the
        # compressed buffer.
        while len(self._uncompressedData) > 0:
            compressed_chunk += self._compress_token_sequence()

        # After we are done, we need to ensure the final compressed chunk is
        # not larger than the original to the point that it exceeds 4096 bytes.
        # If it has, we can copy the original 4096 bytes of uncompressed data
        # and flag this chunk as using "raw" compression by setting the
        # compression but to 0.
        # This chunk size is three less then the "total" chunk size, which is
        # data plus the two byte header.
        chunksize_minus_three = len(compressed_chunk) - 1

        if chunksize_minus_three > 4095:
            chunksize_minus_three = 4095
            # Raw chunks must be 4096 bytes in size. Even if the starting data
            # is less than 4096 bytes. This means after decompressing, there
            # may be unexpected padding.
            compressed_chunk = data.ljust(4096, b'\x00')
            compress_and_sig = 0x3000

        # Join the 12 bit chunk size with the compress bit and three bit
        # siganture.
        header = compress_and_sig | chunksize_minus_three

        # Prepend the header to the compressed chunk data and return the
        # complete chunk.
        compressed_chunk = header.to_bytes(2, self._endian) + compressed_chunk
        return compressed_chunk

    def _compress_token_sequence(self: T) -> bytes:
        """
        A token sequence is a 1 byte token flag followed by 8 tokens. Each
        tokenis one or two bytes. Each bit of the flag byte incates if the
        token is a 1 byte literal token or a 2 byte copy token. The least
        significant bit of the flag byte refers to the first token in the
        stream.
        """

        token_flag = 0
        tokens = b''
        for i in range(8):
            if len(self._uncompressedData) > 0:
                packed_token, flag = self._compress_token()

                # The flag is a 1 or 0, left shift it and pack it into the flag
                # byte.
                token_flag = (flag << i) | token_flag
                tokens += packed_token
        token_sequence = token_flag.to_bytes(1, "little") + tokens
        return token_sequence

    def _compress_token(self: T) -> tuple:
        """
        Given a sequence of uncompressed data, return a single compressed
        token. Tokens are either one byte representing the value of the
        token or two bytes indicating the location and length of the
        replacement sequence. The flag byte is 1 if replacement took place.
        The token that is returned will be packed into the required little
        endian packing.
        Should the flag bit be returned as bool instead of int?
        :return: (packed copy token, flag bit)
        :rtype: (bytes, int)
        """
        packed_token = b''
        token_flag = 0
        offset, length = self._matching()
        if offset > 0:
            token_flag = 1

            # Pack the offset and length data into the CopyToken, then pack
            # the token little-endian.
            difference = len(self._activeChunk) - len(self._uncompressedData)
            help = MsOvba.copytoken_help(difference)
            token_int = MsOvba.pack_copytoken(length, offset, help)
            packed_token = token_int.to_bytes(2, "little")

            # Update the uncompressed buffer by removing the length we were
            # able to match.
            self._uncompressedData = self._uncompressedData[length:]
        else:
            token_flag = 0

            # No matching data was found, copy the literal token over.
            packed_token = self._uncompressedData[0].to_bytes(1, "little")
            self._uncompressedData = self._uncompressedData[1:]
        return packed_token, token_flag

    def _matching(self: T) -> tuple:
        """
        Work backwards through the uncompressed data that has already been
        compressed to find the longest series of matching bytes.
        :return: (offset, length)
        :rtype: (int, int)
        """
        offset = 0
        length = 0
        best_length = 0
        best_candidate = 0
        candidate = len(self._activeChunk) - len(self._uncompressedData) - 1
        while candidate >= 0:
            c = candidate
            d = len(self._activeChunk) - len(self._uncompressedData)
            length = 0
            while (d < len(self._activeChunk)
                   and self._activeChunk[d] == self._activeChunk[c]):
                c += 1
                d += 1
                length += 1
            if length > best_length:
                best_length = length
                best_candidate = candidate
            candidate -= 1

        if best_length >= 3:
            difference = len(self._activeChunk) - len(self._uncompressedData)
            help = MsOvba.copytoken_help(difference)
            maximum_length = help["maxLength"]
            length = min(maximum_length, best_length)
            offset = (len(self._activeChunk) - len(self._uncompressedData)
                      - best_candidate)

        return offset, length

    @staticmethod
    def copytoken_help(difference: int) -> dict:
        """
        Calculate a lengthMask, offsetMask, and bitCount from the length of the
        uncompressedData.
        """
        bit_count = MsOvba.ceil_log2(difference)
        length_mask = 0xFFFF >> bit_count
        offset_mask = ~length_mask & 0xFFFF
        max_length = 0xFFFF << bit_count + 3
        return {
            "lengthMask": length_mask,
            "offsetMask": offset_mask,
            "bitCount": bit_count,
            "maxLength": max_length
        }

    @staticmethod
    def unpack_copytoken(copytoken: int, help: dict) -> dict:
        """
        calculate an offset and length from a 16 bit copytoken
        """
        length = (copytoken & help["lengthMask"]) + 3
        temp1 = copytoken & help["offsetMask"]
        temp2 = 16 - help["bitCount"]
        offset = (temp1 >> temp2) + 1
        return {
            "length": length,
            "offset": offset
        }

    @staticmethod
    def pack_copytoken(length: int, offset: int, help: dict) -> int:
        """
        Create the copy token from the length, offset, and currect position
        return bytes
        """
        temp1 = offset - 1
        temp2 = 16 - help["bitCount"]
        temp3 = length - 3
        return (temp1 << temp2) | temp3

    @staticmethod
    def ceil_log2(int: int) -> int:
        i = 4
        while 2 ** i < int:
            i += 1
        return i
